@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 '''
 benchmark.py - this file is part of S3QL (http://s3ql.googlecode.com)
 
@@ -11,10 +11,8 @@ Copyright (C) 2010 Nikolaus Rath <Nikolaus@rath.org>
 This program can be distributed under the terms of the GNU GPLv3.
 '''
 
-from __future__ import division, print_function, absolute_import
 import argparse
 import atexit
-import logging
 import os
 import shutil
 import subprocess
@@ -28,15 +26,19 @@ basedir = os.path.abspath(os.path.join(os.path.dirname(sys.argv[0]), '..'))
 if (os.path.exists(os.path.join(basedir, 'setup.py')) and
     os.path.exists(os.path.join(basedir, 'src', 's3ql', '__init__.py'))):
     sys.path = [os.path.join(basedir, 'src')] + sys.path
+    exec_prefix = os.path.join(basedir, 'bin', '')
+else:
+    exec_prefix = ''
 
+from s3ql.logging import logging, setup_logging, QuietError
 from s3ql.backends.common import get_backend, BetterBackend, DanglingStorageURLError
 from s3ql.backends.local import Backend
-from s3ql.common import setup_logging, BUFSIZE, QuietError
+from s3ql.common import BUFSIZE
 from s3ql.parse_args import ArgumentParser
 
 ALGS = ('lzma', 'bzip2', 'zlib')
 
-log = logging.getLogger('benchmark')
+log = logging.getLogger(__name__)
 
 def parse_args(args):
     '''Parse command line'''
@@ -78,8 +80,8 @@ def main(args=None):
             copied += len(buf)
             
     log.info('Measuring throughput to cache...')
-    backend_dir = tempfile.mkdtemp()
-    mnt_dir = tempfile.mkdtemp()
+    backend_dir = tempfile.mkdtemp(prefix='s3ql-benchmark-')
+    mnt_dir = tempfile.mkdtemp(prefix='s3ql-mnt')
     atexit.register(shutil.rmtree, backend_dir)
     atexit.register(shutil.rmtree, mnt_dir)
     
@@ -87,9 +89,9 @@ def main(args=None):
     size = 50 * 1024 * 1024
     while write_time < 3:        
         log.debug('Write took %.3g seconds, retrying', write_time) 
-        subprocess.check_call(['mkfs.s3ql', '--plain', 'local://%s' % backend_dir,
+        subprocess.check_call([exec_prefix + 'mkfs.s3ql', '--plain', 'local://%s' % backend_dir,
                                '--quiet', '--force', '--cachedir', options.cachedir])
-        subprocess.check_call(['mount.s3ql', '--threads', '1', '--quiet',
+        subprocess.check_call([exec_prefix + 'mount.s3ql', '--threads', '1', '--quiet',
                                '--cachesize', '%d' % (2 * size / 1024), '--log',
                                '%s/mount.log' % backend_dir, '--cachedir', options.cachedir,
                                'local://%s' % backend_dir, mnt_dir])
@@ -110,7 +112,7 @@ def main(args=None):
             write_time = time.time() - write_time
             os.unlink('%s/bigfile' % mnt_dir)
         finally:
-            subprocess.check_call(['umount.s3ql', mnt_dir])
+            subprocess.check_call([exec_prefix + 'umount.s3ql', mnt_dir])
             
     fuse_speed = copied / write_time
     log.info('Cache throughput: %d KiB/sec', fuse_speed / 1024)
@@ -121,7 +123,7 @@ def main(args=None):
     try:
         backend = get_backend(options, plain=True)
     except DanglingStorageURLError as exc:
-        raise QuietError(str(exc))
+        raise QuietError(str(exc)) from None
     
     upload_time = 0
     size = 512 * 1024
@@ -152,8 +154,8 @@ def main(args=None):
     in_speed = dict()
     out_speed = dict()
     for alg in ALGS:
-        log.info('compressing with %s...', alg)
-        backend = BetterBackend('pass', alg, Backend('local://' + backend_dir, None, None))
+        log.info('compressing with %s-6...', alg)
+        backend = BetterBackend(b'pass', (alg, 6), Backend('local://' + backend_dir, None, None))
         def do_write(dst): #pylint: disable=E0102
             src.seek(0)
             stamp = time.time()
