@@ -6,36 +6,34 @@ Copyright (C) Nikolaus Rath <Nikolaus@rath.org>
 This program can be distributed under the terms of the GNU GPLv3.
 '''
 
-from __future__ import division, print_function, absolute_import
-from ..common import BUFSIZE, QuietError
-from .common import AbstractBackend, NoSuchObject, retry, AuthorizationError, http_connection, \
-    AuthenticationError
-from .common import DanglingStorageURLError
-from base64 import b64encode
+from ..logging import logging # Ensure use of custom logger class
+from ..common import BUFSIZE, QuietError, PICKLE_PROTOCOL, ChecksumError, md5sum
+from .common import (AbstractBackend, NoSuchObject, retry, AuthorizationError, http_connection, 
+    AuthenticationError, DanglingStorageURLError, is_temp_network_error, retry_generator)
+from ..inherit_docstrings import (copy_ancestor_docstring, prepend_ancestor_docstring,
+                                  ABCDocstMeta)
+from base64 import b64encode, b64decode
 from email.utils import parsedate_tz, mktime_tz
-from urlparse import urlsplit
-from httplib import _MAXLINE, CONTINUE, LineTooLong
+from urllib.parse import urlsplit
+from xml.etree import ElementTree
+from http.client import _MAXLINE, CONTINUE, LineTooLong
 import hashlib
 import hmac
-import httplib
-import logging
-import sys
+import http.client
 import re
 import tempfile
 import time
-import urllib
-import xml.etree.cElementTree as ElementTree
-from s3ql.backends.common import is_temp_network_error
-
+import urllib.parse
+import pickle
 
 C_DAY_NAMES = [ 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun' ]
 C_MONTH_NAMES = [ 'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec' ]
 
 XML_CONTENT_RE = re.compile(r'^(?:application|text)/xml(?:;|$)', re.IGNORECASE)
 
-log = logging.getLogger("backends.s3c")
+log = logging.getLogger(__name__)
 
-class Backend(AbstractBackend):
+class Backend(AbstractBackend, metaclass=ABCDocstMeta):
     """A backend to stored data in some S3 compatible storage service.
     
     This class uses standard HTTP connections to connect to GS.
@@ -48,24 +46,39 @@ class Backend(AbstractBackend):
 
     use_expect_100c = True
 
-    def __init__(self, storage_url, login, password, use_ssl):
-        super(Backend, self).__init__()
+    def __init__(self, storage_url, login, password, ssl_context):
+        '''Initialize backend object
 
-        (host, port, bucket_name, prefix) = self._parse_storage_url(storage_url, use_ssl)
+        *ssl_context* may be a `ssl.SSLContext` instance or *None*.
+        '''
+        
+        super().__init__()
+
+        (host, port, bucket_name, prefix) = self._parse_storage_url(storage_url, ssl_context)
 
         self.bucket_name = bucket_name
         self.prefix = prefix
         self.hostname = host
         self.port = port
-        self.use_ssl = use_ssl
+        self.ssl_context = ssl_context
         self.conn = self._get_conn()
 
         self.password = password
         self.login = login
-        self.namespace = 'http://s3.amazonaws.com/doc/2006-03-01/'
+        self.xml_ns_prefix = '{http://s3.amazonaws.com/doc/2006-03-01/}'
+
+    @copy_ancestor_docstring
+    def reset(self):
+        # A bit of a nasty hack, but a better solution that fully
+        # replaces the annoying http.client class is already in the works...
+        if (self.conn._HTTPConnection__state != http.client._CS_IDLE
+            or (self.conn._HTTPConnection__response
+                and not self.conn._HTTPConnection__response.isclosed())):
+            log.debug('Resetting state of http connection %d', id(self.conn))
+            self.conn.close()
 
     @staticmethod
-    def _parse_storage_url(storage_url, use_ssl):
+    def _parse_storage_url(storage_url, ssl_context):
         '''Extract information from storage URL
         
         Return a tuple * (host, port, bucket_name, prefix) * .
@@ -83,7 +96,7 @@ class Backend(AbstractBackend):
         hostname = hit.group(1)
         if hit.group(2):
             port = int(hit.group(2))
-        elif use_ssl:
+        elif ssl_context:
             port = 443
         else:
             port = 80
@@ -95,20 +108,10 @@ class Backend(AbstractBackend):
     def _get_conn(self):
         '''Return connection to server'''
         
-        return http_connection(self.hostname, self.port, self.use_ssl)
+        return http_connection(self.hostname, self.port, self.ssl_context)
 
+    @copy_ancestor_docstring
     def is_temp_failure(self, exc): #IGNORE:W0613
-        '''Return true if exc indicates a temporary error
-    
-        Return true if the given exception indicates a temporary problem. Most instance methods
-        automatically retry the request in this case, so the caller does not need to worry about
-        temporary failures.
-        
-        However, in same cases (e.g. when reading or writing an object), the request cannot
-        automatically be retried. In these case this method can be used to check for temporary
-        problems and so that the request can be manually restarted if applicable.
-        '''
-
         if isinstance(exc, (InternalError, BadDigestError, IncompleteBodyError, 
                             RequestTimeoutError, OperationAbortedError, SlowDownError, 
                             RequestTimeTooSkewedError)):
@@ -123,9 +126,8 @@ class Backend(AbstractBackend):
         return False
 
     @retry
+    @copy_ancestor_docstring
     def delete(self, key, force=False):
-        '''Delete the specified object'''
-
         log.debug('delete(%s)', key)
         try:
             resp = self._do_request('DELETE', '/%s%s' % (self.prefix, key))
@@ -134,59 +136,17 @@ class Backend(AbstractBackend):
             if force:
                 pass
             else:
-                raise NoSuchObject(key)
+                raise NoSuchObject(key) from None
 
-    def list(self, prefix=''):
-        '''List keys in backend
-
-        Returns an iterator over all keys in the backend. This method
-        handles temporary errors.
-        '''
-
-        log.debug('list(%s): start', prefix)
-
-        marker = ''
-        waited = 0
-        interval = 1 / 50
-        iterator = self._list(prefix, marker)
-        while True:
-            try:
-                marker = iterator.next()
-                waited = 0
-            except StopIteration:
-                break
-            except Exception as exc:
-                if not self.is_temp_failure(exc):
-                    raise
-                if waited > 60 * 60:
-                    log.error('list(): Timeout exceeded, re-raising %s exception', 
-                              type(exc).__name__)
-                    raise
-
-                log.info('Encountered %s exception (%s), retrying call to s3c.Backend.list()',
-                          type(exc).__name__, exc)
-                
-                if hasattr(exc, 'retry_after') and exc.retry_after:
-                    interval = exc.retry_after
-                                    
-                time.sleep(interval)
-                waited += interval
-                interval = min(5*60, 2*interval)
-                iterator = self._list(prefix, marker)
-
-            else:
-                yield marker
-
-    def _list(self, prefix='', start=''):
-        '''List keys in backend, starting with *start*
-
-        Returns an iterator over all keys in the backend. This method
-        does not retry on errors.
-        '''
+    @retry_generator
+    @copy_ancestor_docstring
+    def list(self, prefix='', start_after=''):
+        log.debug('list(%s, %s): start', prefix, start_after)
 
         keys_remaining = True
-        marker = start
+        marker = start_after
         prefix = self.prefix + prefix
+        ns_p = self.xml_ns_prefix
 
         while keys_remaining:
             log.debug('list(%s): requesting with marker=%s', prefix, marker)
@@ -200,22 +160,18 @@ class Backend(AbstractBackend):
                 raise RuntimeError('unexpected content type: %s' % resp.getheader('Content-Type'))
 
             itree = iter(ElementTree.iterparse(resp, events=("start", "end")))
-            (event, root) = itree.next()
-
-            namespace = re.sub(r'^\{(.+)\}.+$', r'\1', root.tag)
-            if namespace != self.namespace:
-                raise RuntimeError('Unsupported namespace: %s' % namespace)
+            (event, root) = next(itree)
 
             try:
                 for (event, el) in itree:
                     if event != 'end':
                         continue
 
-                    if el.tag == '{%s}IsTruncated' % self.namespace:
+                    if el.tag == ns_p + 'IsTruncated':
                         keys_remaining = (el.text == 'true')
 
-                    elif el.tag == '{%s}Contents' % self.namespace:
-                        marker = el.findtext('{%s}Key' % self.namespace)
+                    elif el.tag == ns_p + 'Contents':
+                        marker = el.findtext(ns_p + 'Key')
                         yield marker[len(self.prefix):]
                         root.clear()
 
@@ -223,7 +179,7 @@ class Backend(AbstractBackend):
                 # Need to read rest of response
                 while True:
                     buf = resp.read(BUFSIZE)
-                    if buf == '':
+                    if buf == b'':
                         break
                 break
 
@@ -231,9 +187,8 @@ class Backend(AbstractBackend):
                 raise RuntimeError('Could not parse body')
 
     @retry
+    @copy_ancestor_docstring
     def lookup(self, key):
-        """Return metadata for given key"""
-
         log.debug('lookup(%s)', key)
 
         try:
@@ -241,16 +196,15 @@ class Backend(AbstractBackend):
             assert resp.length == 0
         except HTTPError as exc:
             if exc.status == 404:
-                raise NoSuchObject(key)
+                raise NoSuchObject(key) from None
             else:
                 raise
 
         return extractmeta(resp)
 
     @retry
+    @copy_ancestor_docstring
     def get_size(self, key):
-        '''Return size of object stored under *key*'''
-
         log.debug('get_size(%s)', key)
 
         try:
@@ -258,7 +212,7 @@ class Backend(AbstractBackend):
             assert resp.length == 0
         except HTTPError as exc:
             if exc.status == 404:
-                raise NoSuchObject(key)
+                raise NoSuchObject(key) from None
             else:
                 raise
 
@@ -269,54 +223,43 @@ class Backend(AbstractBackend):
 
 
     @retry
+    @copy_ancestor_docstring
     def open_read(self, key):
-        """Open object for reading
-
-        Return a file-like object. Data can be read using the `read` method. metadata is stored in
-        its *metadata* attribute and can be modified by the caller at will. The object must be
-        closed explicitly.
-        """
-
         try:
             resp = self._do_request('GET', '/%s%s' % (self.prefix, key))
         except NoSuchKeyError:
-            raise NoSuchObject(key)
+            raise NoSuchObject(key) from None
 
         return ObjectR(key, resp, self, extractmeta(resp))
 
+    @prepend_ancestor_docstring
     def open_write(self, key, metadata=None, is_compressed=False):
-        """Open object for writing
-
-        `metadata` can be a dict of additional attributes to store with the object. Returns a file-
-        like object. The object must be closed explicitly. After closing, the *get_obj_size* may be
-        used to retrieve the size of the stored object (which may differ from the size of the
-        written data).
-        
-        The *is_compressed* parameter indicates that the caller is going to write compressed data,
-        and may be used to avoid recompression by the backend.
-                        
-        Since Amazon S3 does not support chunked uploads, the entire data will
-        be buffered in memory before upload.
+        """
+        The returned object will buffer all data and only start the upload
+        when its `close` method is called.
         """
 
         log.debug('open_write(%s): start', key)
 
+        # We don't store the metadata keys directly, because HTTP headers
+        # are case insensitive (so the server may change capitalization)
+        # and we may run into length restrictions.
+        meta_buf = b64encode(pickle.dumps(metadata, PICKLE_PROTOCOL)).decode('us-ascii')
+
+        chunksize = 255
+        i = 0
         headers = dict()
-        if metadata:
-            for (hdr, val) in metadata.iteritems():
-                headers['x-amz-meta-%s' % hdr] = val
+        headers['x-amz-meta-format'] = 'pickle'
+        while i*chunksize < len(meta_buf):
+            headers['x-amz-meta-data-%02d' % i] = meta_buf[i*chunksize:(i+1)*chunksize]
+            i += 1
 
         return ObjectW(key, self, headers)
 
 
     @retry
+    @copy_ancestor_docstring
     def copy(self, src, dest):
-        """Copy data stored under key `src` to key `dest`
-        
-        If `dest` already exists, it will be overwritten. The copying is done on
-        the remote side.
-        """
-
         log.debug('copy(%s, %s): start', src, dest)
 
         try:
@@ -326,7 +269,7 @@ class Backend(AbstractBackend):
             # Discard response body
             resp.read()
         except NoSuchKeyError:
-            raise NoSuchObject(src)
+            raise NoSuchObject(src) from None
 
     def _do_request(self, method, path, subres=None, query_string=None,
                     headers=None, body=None):
@@ -342,46 +285,69 @@ class Backend(AbstractBackend):
 
         if not body:
             headers['content-length'] = '0'
+        elif isinstance(body, (str, bytes, bytearray, memoryview)):
+            if isinstance(body, str):
+                body = body.encode('ascii')
+            headers['content-length'] = '%d' % (len(body))
+            headers['content-md5'] = b64encode(md5sum(body)).decode('ascii')
 
         redirect_count = 0
         while True:
                 
             resp = self._send_request(method, path, headers, subres, query_string, body)
-            log.debug('_do_request(): request-id: %s', resp.getheader('x-amz-request-id'))
+            log.debug('status: %d, request-id: %s', resp.status, resp.getheader('x-amz-request-id'))
 
             if (resp.status < 300 or resp.status > 399):
                 break
 
             # Assume redirect
-            new_url = resp.getheader('Location')
-            if new_url is None:
-                break
-            log.info('_do_request(): redirected to %s', new_url)
-                        
             redirect_count += 1
             if redirect_count > 10:
                 raise RuntimeError('Too many chained redirections')
-    
-            # Pylint can't infer SplitResult Types
-            #pylint: disable=E1103
-            o = urlsplit(new_url)
-            if o.scheme:
-                if isinstance(self.conn, httplib.HTTPConnection) and o.scheme != 'http':
-                    raise RuntimeError('Redirect to non-http URL')
-                elif isinstance(self.conn, httplib.HTTPSConnection) and o.scheme != 'https':
-                    raise RuntimeError('Redirect to non-https URL')
-            if o.hostname != self.hostname or o.port != self.port:
-                self.hostname = o.hostname
-                self.port = o.port
-                self.conn = self._get_conn()
-            else:
-                raise RuntimeError('Redirect to different path on same host')
-            
-            if body and not isinstance(body, bytes):
-                body.seek(0)
 
-            # Read and discard body
-            log.debug('Response body: %s', resp.read())
+            # First try location header...
+            new_url = resp.getheader('Location')
+            if new_url:
+                # Read and discard body
+                while True:
+                    if not resp.read(BUFSIZE):
+                        break
+
+                # Pylint can't infer SplitResult Types
+                #pylint: disable=E1103
+                o = urlsplit(new_url)
+                if o.scheme:
+                    if self.ssl_context and o.scheme != 'https':
+                        raise RuntimeError('Redirect to non-https URL')
+                    elif not self.ssl_context and o.scheme != 'http':
+                        raise RuntimeError('Redirect to non-http URL')
+                if o.hostname != self.hostname or o.port != self.port:
+                    self.hostname = o.hostname
+                    self.port = o.port
+                    self.conn = self._get_conn()
+                else:
+                    raise RuntimeError('Redirect to different path on same host')
+                
+            # ..but endpoint may also be hidden in message body..
+            else:
+                # If we got a HEAD response, the server cannot transmit the
+                # request body with the endpoint
+                if method == 'HEAD':
+                    raise HTTPError(resp.status, resp.reason)
+
+                tree = self._parse_xml_response(resp)
+                new_url = tree.findtext('Endpoint')
+
+                if not new_url:
+                    raise get_S3Error(tree.findtext('Code'), tree.findtext('Message'))
+
+                self.hostname = new_url
+                self.conn = self._get_conn()
+                
+            log.info('_do_request(): redirected to %s', new_url)
+
+            if body and not isinstance(body, (bytes, bytearray, memoryview)):
+                body.seek(0)
 
         # We need to call read() at least once for httplib to consider this
         # request finished, even if there is no response body.
@@ -397,20 +363,38 @@ class Backend(AbstractBackend):
         if method.upper() == 'HEAD':
             raise HTTPError(resp.status, resp.reason)
 
-        content_type = resp.getheader('Content-Type')
-        if not content_type or not XML_CONTENT_RE.match(content_type):
-            raise HTTPError(resp.status, resp.reason, resp.getheaders(), resp.read())
- 
         # Error
-        tree = ElementTree.parse(resp).getroot()
+        tree = self._parse_xml_response(resp)
         raise get_S3Error(tree.findtext('Code'), tree.findtext('Message'))
 
+    
+    def _parse_xml_response(self, resp):
+        '''Return element tree for XML response'''
 
+        content_type = resp.getheader('Content-Type')
+
+        # AWS S3 sometimes "forgets" to send a Content-Type
+        # when responding to a multiple delete request.
+        # https://forums.aws.amazon.com/thread.jspa?threadID=134372
+        if content_type is None:
+            log.error('Server did not provide Content-Type, assuming XML')
+        elif not XML_CONTENT_RE.match(content_type):
+            log.error('Unexpected server reply: expected XML, got %r', content_type)
+            log.error('Full response:\n%d %s\n%s\n\n%s', resp.status, resp.reason,
+                      '\n'.join('%s: %s' % x for x in resp.getheaders()),
+                      resp.read(512))
+            raise RuntimeError('Unexpected server response')
+ 
+        tree = ElementTree.parse(resp).getroot()
+
+        return tree
+
+    
+    @prepend_ancestor_docstring
     def clear(self):
-        """Delete all objects in backend
-        
-        Note that this method may not be able to see (and therefore also not
-        delete) recently uploaded objects.
+        """
+        This method may not be able to see (and therefore also not delete)
+        recently uploaded objects.
         """
 
         # We have to cache keys, because otherwise we can't use the
@@ -430,19 +414,13 @@ class Backend(AbstractBackend):
     def _send_request(self, method, path, headers, subres=None, query_string=None, body=None):
         '''Add authentication and send request
         
-        Note that *headers* is modified in-place. Returns the response object.
+        Returns the response object.
         '''
 
         # See http://docs.amazonwebservices.com/AmazonS3/latest/dev/RESTAuthentication.html
 
         # Lowercase headers
-        keys = list(headers.iterkeys())
-        for key in keys:
-            key_l = key.lower()
-            if key_l == key:
-                continue
-            headers[key_l] = headers[key]
-            del headers[key]
+        headers = { x.lower(): y for (x,y) in headers.items() }
 
         # Date, can't use strftime because it's locale dependent
         now = time.gmtime()
@@ -466,23 +444,25 @@ class Backend(AbstractBackend):
 
 
         # Always include bucket name in path for signing
-        sign_path = urllib.quote('/%s%s' % (self.bucket_name, path))
+        sign_path = urllib.parse.quote('/%s%s' % (self.bucket_name, path))
         auth_strs.append(sign_path)
         if subres:
             auth_strs.append('?%s' % subres)
 
         # False positive, hashlib *does* have sha1 member
         #pylint: disable=E1101
-        signature = b64encode(hmac.new(self.password, ''.join(auth_strs), hashlib.sha1).digest())
+        auth_str = ''.join(auth_strs).encode()
+        signature = b64encode(hmac.new(self.password.encode(), auth_str,
+                                       hashlib.sha1).digest()).decode()
 
         headers['authorization'] = 'AWS %s:%s' % (self.login, signature)
-
+        
         # Construct full path
         if not self.hostname.startswith(self.bucket_name):
             path = '/%s%s' % (self.bucket_name, path)
-        path = urllib.quote(path)
+        path = urllib.parse.quote(path)
         if query_string:
-            s = urllib.urlencode(query_string, doseq=True)
+            s = urllib.parse.urlencode(query_string, doseq=True)
             if subres:
                 path += '?%s&%s' % (subres, s)
             else:
@@ -511,20 +491,25 @@ class Backend(AbstractBackend):
             # so that it knows about the body that still needs to
             # be sent...
             resp = HTTPResponse(self.conn.sock, body)
+            native_response_class = self.conn.response_class
             try:
-                native_response_class = self.conn.response_class
                 self.conn.response_class = resp
                 assert self.conn.getresponse() is resp
             finally:
                 self.conn.response_class = native_response_class
             return resp
             
-        except:
-            # We probably can't use the connection anymore
-            self.conn.close()
+        except Exception as exc:
+            if is_temp_network_error(exc):
+                # We probably can't use the connection anymore
+                self.conn.close()
             raise
+
+    @copy_ancestor_docstring
+    def close(self):
+        self.conn.close()
         
-class HTTPResponse(httplib.HTTPResponse):
+class HTTPResponse(http.client.HTTPResponse):
     '''
     This class provides a HTTP Response object that supports waiting for
     "100 Continue" and then sending the request body.
@@ -551,18 +536,18 @@ class HTTPResponse(httplib.HTTPResponse):
         self.body = body
         self.__cached_status = None
         
-    def __call__(self, sock, *a, **kw):
+    def __call__(self, sock, debuglevel=0, method=None, url=None):
         '''Fake object instantiation'''
 
         assert self.sock is sock
 
-        httplib.HTTPResponse.__init__(self, sock, *a, **kw)
+        super().__init__(sock, debuglevel, method=method, url=url)
 
         return self
 
     def _read_status(self):
         if self.__cached_status is None:
-            self.__cached_status = httplib.HTTPResponse._read_status(self)
+            self.__cached_status = super()._read_status()
 
         return self.__cached_status
 
@@ -572,7 +557,7 @@ class HTTPResponse(httplib.HTTPResponse):
         (version, status, reason) = self._read_status()
         if status != CONTINUE:
             # Oops, error. Let regular code take over
-            return httplib.HTTPResponse.begin(self)
+            return super().begin()
 
         # skip the header from the 100 response
         while True:
@@ -596,12 +581,16 @@ class HTTPResponse(httplib.HTTPResponse):
 
         # *Now* read the actual response
         self.__cached_status = None
-        return httplib.HTTPResponse.begin(self)
+        return super().begin()
 
                 
 class ObjectR(object):
     '''An S3 object open for reading'''
 
+    # NOTE: This class is used as a base class for the swift backend,
+    # so changes here should be checked for their effects on other
+    # backends.
+    
     def __init__(self, key, resp, backend, metadata=None):
         self.key = key
         self.resp = resp
@@ -630,25 +619,20 @@ class ObjectR(object):
             self.md5_checked = True
             
             if not self.resp.isclosed():
-                # http://bugs.python.org/issue15633
-                ver = sys.version_info
-                if ((ver > (2,7,3) and ver < (3,0,0))
-                    or (ver > (3,2,3) and ver < (3,3,0))
-                    or (ver > (3,3,0))):
-                    # Should be fixed in these version
-                    log.error('ObjectR.read(): response not closed after end of data, '
-                              'please report on http://code.google.com/p/s3ql/issues/')
-                    log.error('Method: %s, chunked: %s, read length: %s '
-                              'response length: %s, chunk_left: %s, status: %d '
-                              'reason "%s", version: %s, will_close: %s',
-                              self.resp._method, self.resp.chunked, size, self.resp.length,
-                              self.resp.chunk_left, self.resp.status, self.resp.reason,
-                              self.resp.version, self.resp.will_close)
-                                    
+                # http://bugs.python.org/issue15633, but should be fixed in 
+                # Python 3.3 and newer
+                log.error('ObjectR.read(): response not closed after end of data, '
+                          'please report on http://code.google.com/p/s3ql/issues/')
+                log.error('Method: %s, chunked: %s, read length: %s '
+                          'response length: %s, chunk_left: %s, status: %d '
+                          'reason "%s", version: %s, will_close: %s',
+                          self.resp._method, self.resp.chunked, size, self.resp.length,
+                          self.resp.chunk_left, self.resp.status, self.resp.reason,
+                          self.resp.version, self.resp.will_close)             
                 self.resp.close() 
             
             if etag != self.md5.hexdigest():
-                log.warn('ObjectR(%s).close(): MD5 mismatch: %s vs %s', self.key, etag,
+                log.warning('ObjectR(%s).close(): MD5 mismatch: %s vs %s', self.key, etag,
                          self.md5.hexdigest())
                 raise BadDigestError('BadDigest', 'ETag header does not agree with calculated MD5')
             
@@ -674,14 +658,22 @@ class ObjectW(object):
     All data is first cached in memory, upload only starts when
     the close() method is called.
     '''
-
+    
+    # NOTE: This class is used as a base class for the swift backend,
+    # so changes here should be checked for their effects on other
+    # backends.
+    
     def __init__(self, key, backend, headers):
         self.key = key
         self.backend = backend
         self.headers = headers
         self.closed = False
         self.obj_size = 0
-        self.fh = tempfile.TemporaryFile(bufsize=0) # no Python buffering
+        
+        # According to http://docs.python.org/3/library/functions.html#open
+        # the buffer size is typically ~8 kB. We process data in much 
+        # larger chunks, so buffering would only hurt performance.
+        self.fh = tempfile.TemporaryFile(buffering=0) 
 
         # False positive, hashlib *does* have md5 member
         #pylint: disable=E1101        
@@ -706,24 +698,25 @@ class ObjectW(object):
 
         log.debug('ObjectW(%s).close(): start', self.key)
 
-        self.closed = True
-        self.headers['Content-Length'] = self.obj_size
-
         self.fh.seek(0)
+        self.headers['Content-Length'] = self.obj_size
+        self.headers['Content-Type'] = 'application/octet-stream'
         resp = self.backend._do_request('PUT', '/%s%s' % (self.backend.prefix, self.key),
-                                       headers=self.headers, body=self.fh)
+                                        headers=self.headers, body=self.fh)
         etag = resp.getheader('ETag').strip('"')
         assert resp.length == 0
 
-        if etag != self.md5.hexdigest():
-            log.warn('ObjectW(%s).close(): MD5 mismatch (%s vs %s)', self.key, etag,
-                     self.md5.hexdigest)
-            try:
-                self.backend.delete(self.key)
-            except:
-                log.exception('Objectw(%s).close(): unable to delete corrupted object!',
-                              self.key)
-            raise BadDigestError('BadDigest', 'Received ETag does not agree with our calculations.')
+        # Try-catch to ensure that we don't loose the BadDigest exception
+        try:
+            if etag != self.md5.hexdigest():
+                raise BadDigestError('BadDigest', 'MD5 mismatch for %s (received: %s, sent: %s)' %
+                                     (self.key, etag, self.md5.hexdigest))
+        except:
+            self.backend.delete(self.key)
+            raise
+
+        self.fh.close()
+        self.closed = True
 
     def __enter__(self):
         return self
@@ -741,12 +734,17 @@ class ObjectW(object):
 def get_S3Error(code, msg):
     '''Instantiate most specific S3Error subclass'''
 
+    # Special case
+    # http://code.google.com/p/s3ql/issues/detail?id=369
+    if code == 'Timeout':
+        code = 'RequestTimeout'  
+
     if code.endswith('Error'):
         name = code
     else:
         name = code + 'Error'
-    class_ = globals().get(name, S3Error)
-
+    class_ = globals().get(name, S3Error)    
+    
     if not issubclass(class_, S3Error):
         return S3Error(code, msg)
     
@@ -755,20 +753,36 @@ def get_S3Error(code, msg):
 def extractmeta(resp):
     '''Extract metadata from HTTP response object'''
 
-    # Note: we implicitly rely on httplib to convert all headers to lower
-    # case. because HTTP headers are case sensitive (so meta data field names
-    # may have their capitalization changed). This only works, however, because
-    # the meta data field names that we use are lower case as well. This problem
-    # has only been solved cleanly in S3QL 2.0.
-    
     meta = dict()
+    format_ = 'raw'
     for (name, val) in resp.getheaders():
-        hit = re.match(r'^x-amz-meta-(.+)$', name)
+        # HTTP headers are case-insensitive and pre 2.x S3QL versions metadata
+        # names verbatim (and thus loose capitalization info), so we force lower
+        # case (since we know that this is the original capitalization).
+        name = name.lower()
+
+        hit = re.match(r'^x-amz-meta-(.+)$', name, re.IGNORECASE)
         if not hit:
             continue
-        meta[hit.group(1)] = val
 
-    return meta
+        if hit.group(1).lower() == 'format':
+            format_ = val
+        else:
+            meta[hit.group(1)] = val
+
+    if format_ == 'pickle':
+        buf = ''.join(meta[x] for x in sorted(meta)
+                      if x.startswith('data-'))
+        try:
+            return pickle.loads(b64decode(buf))
+        except pickle.UnpicklingError as exc:
+            if (isinstance(exc.args[0], str)
+                and exc.args[0].startswith('invalid load key')):
+                raise ChecksumError('Invalid metadata') from None
+            raise
+    else:
+        return meta
+
 
 class HTTPError(Exception):
     '''
@@ -776,7 +790,7 @@ class HTTPError(Exception):
     '''
 
     def __init__(self, status, msg, headers=None, body=None):
-        super(HTTPError, self).__init__()
+        super().__init__()
         self.status = status
         self.msg = msg
         self.headers = headers
@@ -798,13 +812,13 @@ class HTTPError(Exception):
                 else:
                     date = parsedate_tz(v)
                     if date is None:
-                        log.warn('Unable to parse header: %s: %s', k, v)
+                        log.warning('Unable to parse header: %s: %s', k, v)
                         continue
                     val = mktime_tz(*date) - time.time()
                     
         if val is not None:
             if val > 300 or val < 0:
-                log.warn('Ignoring invalid retry-after value of %.3f', val)
+                log.warning('Ignoring invalid retry-after value of %.3f', val)
             else:
                 self.retry_after = val
             
@@ -818,7 +832,7 @@ class S3Error(Exception):
     '''
 
     def __init__(self, code, msg):
-        super(S3Error, self).__init__(msg)
+        super().__init__(msg)
         self.code = code
         self.msg = msg
 
@@ -835,7 +849,6 @@ class InvalidSecurityError(S3Error, AuthenticationError): pass
 class SignatureDoesNotMatchError(S3Error, AuthenticationError): pass
 class OperationAbortedError(S3Error): pass
 class RequestTimeoutError(S3Error): pass
-class TimeoutError(RequestTimeoutError): pass
 class SlowDownError(S3Error): pass
 class RequestTimeTooSkewedError(S3Error): pass
 class NoSuchBucketError(S3Error, DanglingStorageURLError): pass
