@@ -6,32 +6,35 @@ Copyright (C) 2008-2009 Nikolaus Rath <Nikolaus@rath.org>
 This program can be distributed under the terms of the GNU GPLv3.
 '''
 
-from __future__ import division, print_function, absolute_import
 from .common import sha256_fh, BUFSIZE
 from .database import NoSuchRowError
-from .ordered_dict import OrderedDict
-from Queue import Queue
+from .logging import logging # Ensure use of custom logger class
+from collections import OrderedDict
 from contextlib import contextmanager
 from llfuse import lock, lock_released
-import logging
+from queue import Queue, Empty as QueueEmtpy
 import os
 import shutil
 import threading
 import time
 
 # standard logger for this module
-log = logging.getLogger("BlockCache")
+log = logging.getLogger(__name__)
 
 # Special queue entry that signals threads to terminate
 QuitSentinel = object()
 
+# Special queue entry that signals that removal queue should
+# be flushed
+FlushSentinel = object()
+
 class Distributor(object):
     '''
     Distributes objects to consumers.
     '''
 
     def __init__(self):
-        super(Distributor, self).__init__()
+        super().__init__()
 
         self.slot = None
         self.cv = threading.Condition()
@@ -75,7 +78,7 @@ class SimpleEvent(object):
     '''
 
     def __init__(self):
-        super(SimpleEvent, self).__init__()
+        super().__init__()
         self.__cond = threading.Condition(threading.Lock())
 
     def notify_all(self):
@@ -122,7 +125,7 @@ class CacheEntry(object):
                   'size', 'pos', 'fh' ]
 
     def __init__(self, inode, blockno, filename):
-        super(CacheEntry, self).__init__()
+        super().__init__()
         # Writing 100MB in 128k chunks takes 90ms unbuffered and
         # 116ms with 1 MB buffer. Reading time does not depend on
         # buffer size.
@@ -221,12 +224,14 @@ class BlockCache(object):
         self.max_size = max_size
         self.in_transit = set()
         self.removed_in_transit = set()
-        self.to_upload = Distributor()
-        self.to_remove = Queue(250)
         self.upload_threads = []
         self.removal_threads = []
         self.transfer_completed = SimpleEvent()
 
+        # Will be initialized once threads are available
+        self.to_upload = None 
+        self.to_remove = None
+        
         if not os.path.exists(self.path):
             os.mkdir(self.path)
 
@@ -237,11 +242,13 @@ class BlockCache(object):
     def init(self, threads=1):
         '''Start worker threads'''
 
+        self.to_upload = Distributor()
         for _ in range(threads):
             t = threading.Thread(target=self._upload_loop)
             t.start()
             self.upload_threads.append(t)
 
+        self.to_remove = Queue(1000)
         for _ in range(10):
             t = threading.Thread(target=self._removal_loop)
             t.daemon = True # interruption will do no permanent harm
@@ -249,31 +256,38 @@ class BlockCache(object):
             self.removal_threads.append(t)
 
     def destroy(self):
-        '''Clean up and stop worker threads'''
+        '''Clean up and stop worker threads
+        
+        This method should be called without the global lock held.
+        '''
 
         log.debug('destroy(): clearing cache...')
-        self.clear()
+        with lock:
+            self.clear()
 
-        with lock_released:
-            for t in self.upload_threads:
-                self.to_upload.put(QuitSentinel)
+        for t in self.upload_threads:
+            self.to_upload.put(QuitSentinel)
 
-            for t in self.removal_threads:
-                self.to_remove.put(QuitSentinel)
+        for t in self.removal_threads:
+            self.to_remove.put(QuitSentinel)
 
-            log.debug('destroy(): waiting for upload threads...')
-            for t in self.upload_threads:
-                t.join()
-
-            log.debug('destroy(): waiting for removal threads...')
-            for t in self.removal_threads:
-                t.join()
-
-        self.upload_threads = []
-        self.removal_threads = []
+        log.debug('destroy(): waiting for upload threads...')
+        for t in self.upload_threads:
+            t.join()
+        self.to_upload = None
+        
+        log.debug('destroy(): waiting for removal threads...')
+        for t in self.removal_threads:
+            t.join()
+        self.to_remove = None
+        
+        self.upload_threads = None
+        self.removal_threads = None
 
         os.rmdir(self.path)
 
+        log.debug('cleanup done.')
+
     def _upload_loop(self):
         '''Process upload queue'''
 
@@ -289,8 +303,13 @@ class BlockCache(object):
         '''Upload object'''
 
         def do_write(fh):
+            # If the object has been removed in transit, the cache element is
+            # already closed and we can't seek.
+            if obj_id in self.removed_in_transit:
+                return fh
+            
             el.seek(0)
-            while True:
+            while obj_id not in self.removed_in_transit:
                 buf = el.read(BUFSIZE)
                 if not buf:
                     break
@@ -304,15 +323,25 @@ class BlockCache(object):
             with self.backend_pool() as backend:
                 obj_size = backend.perform_write(do_write, 's3ql_data_%d' % obj_id).get_obj_size()
 
+            removed = obj_id in self.removed_in_transit
+            
             if log.isEnabledFor(logging.DEBUG):
-                time_ = time.time() - time_
-                rate = el.size / (1024 ** 2 * time_) if time_ != 0 else 0
-                log.debug('_do_upload(%s): uploaded %d bytes in %.3f seconds, %.2f MiB/s',
-                          obj_id, el.size, time_, rate)
+                if removed:
+                    log.debug('_do_upload(%s): aborted upload of deleted object', obj_id)
+                else:
+                    time_ = time.time() - time_
+                    rate = el.size / (1024 ** 2 * time_) if time_ != 0 else 0
+                    log.debug('_do_upload(%s): uploaded %d bytes in %.3f seconds, %.2f MiB/s',
+                              obj_id, el.size, time_, rate)
+
+            if removed:
+                self.removed_in_transit.remove(obj_id)
+                self.to_remove.put(obj_id)
 
             with lock:
-                self.db.execute('UPDATE objects SET size=? WHERE id=?',
-                                (obj_size, obj_id))
+                if not removed:
+                    self.db.execute('UPDATE objects SET size=? WHERE id=?',
+                                    (obj_size, obj_id))
 
                 el.dirty = False
                 self.in_transit.remove(obj_id)
@@ -384,11 +413,7 @@ class BlockCache(object):
 
                 self.in_transit.add(obj_id)
                 with lock_released:
-                    if not self.upload_threads:
-                        log.warn("upload(%s): no upload threads, uploading synchronously", el)
-                        self._do_upload(el, obj_id)
-                    else:
-                        self.to_upload.put((el, obj_id))
+                    self.to_upload.put((el, obj_id))
 
             # There is a block with the same hash                        
             else:
@@ -433,17 +458,13 @@ class BlockCache(object):
         log.debug('upload(%s): removing object %d', el, old_obj_id)
         self.db.execute('DELETE FROM objects WHERE id=?', (old_obj_id,))
 
-        while old_obj_id in self.in_transit:
-            log.debug('upload(%s): waiting for transfer of old object %d to complete',
+        if old_obj_id in self.in_transit:
+            log.debug('upload(%s): deferring removal of old object %d, still in transit',
                       el, old_obj_id)
-            self.wait()
-
-        with lock_released:
-            if not self.removal_threads:
-                log.warn("upload(%s): no removal threads, removing synchronously", el)
-                self._do_removal(old_obj_id)
-            else:
-                log.debug('upload(%s): adding %d to removal queue', el, old_obj_id)
+            self.removed_in_transit.add(old_obj_id)
+        else:
+            log.debug('upload(%s): adding %d to removal queue', el, old_obj_id)
+            with lock_released:
                 self.to_remove.put(old_obj_id)
 
         return el.size
@@ -457,19 +478,25 @@ class BlockCache(object):
     def _removal_loop(self):
         '''Process removal queue'''
 
+        ids = []
         while True:
-            tmp = self.to_remove.get()
+            try:
+                log.debug('reading from queue (blocking=%s)', len(ids)==0)
+                tmp = self.to_remove.get(block=len(ids)==0)
+            except QueueEmtpy:
+                tmp = FlushSentinel
+
+            if (tmp is FlushSentinel or tmp is QuitSentinel) and ids:
+                log.debug('removing: %s', ids)
+                with self.backend_pool() as backend:
+                    backend.delete_multi(['s3ql_data_%d' % i for i in ids])
+                ids = []
+            else:
+                ids.append(tmp)
 
             if tmp is QuitSentinel:
                 break
 
-            self._do_removal(tmp)
-
-    def _do_removal(self, obj_id):
-        '''Remove object'''
-
-        with self.backend_pool() as backend:
-            backend.delete('s3ql_data_%d' % obj_id)
 
     @contextmanager
     def get(self, inode, blockno):
@@ -528,14 +555,17 @@ class BlockCache(object):
                     self.in_transit.add(obj_id)
                     log.debug('get(inode=%d, block=%d): downloading object %d..',
                               inode, blockno, obj_id)
-                    def do_read(fh):
-                        el = CacheEntry(inode, blockno, filename)
-                        shutil.copyfileobj(fh, el, BUFSIZE)
-                        return el
+                        
+                    el = CacheEntry(inode, blockno, filename)    
                     try:
+                        def do_read(fh):
+                            el.seek(0)
+                            el.truncate()
+                            shutil.copyfileobj(fh, el, BUFSIZE)
+
                         with lock_released:
                             with self.backend_pool() as backend:
-                                el = backend.perform_read(do_read, 's3ql_data_%d' % obj_id)
+                                backend.perform_read(do_read, 's3ql_data_%d' % obj_id)
 
                         # Note: We need to do this *before* releasing the global
                         # lock to notify other threads
@@ -546,8 +576,8 @@ class BlockCache(object):
                         self.size += el.size
 
                     except:
-                        if el is not None:
-                            el.unlink()
+                        el.unlink()
+                        el.close()
                         raise
                     finally:
                         self.in_transit.remove(obj_id)
@@ -557,7 +587,7 @@ class BlockCache(object):
             # In Cache
             else:
                 #log.debug('get(inode=%d, block=%d): in cache', inode, blockno)
-                self.entries.to_head((inode, blockno))
+                self.entries.move_to_end((inode, blockno), last=True) # move to head
 
         el.last_access = time.time()
         oldsize = el.size
@@ -592,7 +622,7 @@ class BlockCache(object):
             need_entries = len(self.entries) - self.max_entries
 
             # Try to expire entries that are not dirty
-            for el in self.entries.values_rev():
+            for el in self.entries.values():
                 if el.dirty:
                     if (el.inode, el.blockno) in self.in_transit:
                         log.debug('expire: %s is dirty, but already being uploaded', el)
@@ -616,7 +646,7 @@ class BlockCache(object):
                 break
 
             # Try to upload just enough
-            for el in self.entries.values_rev():
+            for el in self.entries.values():
                 if el.dirty and (el.inode, el.blockno) not in self.in_transit:
                     log.debug('expire: uploading %s..', el)
                     freed = self.upload(el) # Releases global lock
@@ -662,8 +692,6 @@ class BlockCache(object):
             end_no = start_no + 1
 
         for blockno in range(start_no, end_no):
-            # We can't use self.mlock here to prevent simultaneous retrieval
-            # of the block with get(), because this could deadlock
             if (inode, blockno) in self.entries:
                 log.debug('remove(inode=%d, blockno=%d): removing from cache',
                           inode, blockno)
@@ -674,6 +702,7 @@ class BlockCache(object):
 
                 self.size -= el.size
                 el.unlink()
+                el.close()
 
             try:
                 block_id = self.db.get_val('SELECT block_id FROM inode_blocks '
@@ -709,19 +738,16 @@ class BlockCache(object):
                 self.db.execute('UPDATE objects SET refcount=refcount-1 WHERE id=?',
                                 (obj_id,))
             else:
-                while obj_id in self.in_transit:
-                    log.debug('remove(inode=%d, blockno=%d): waiting for transfer of '
-                              'object %d to complete', inode, blockno, obj_id)
-                    self.wait()
-                log.debug('remove(inode=%d, blockno=%d): deleting object %d',
-                          inode, blockno, obj_id)
-                self.db.execute('DELETE FROM objects WHERE id=?', (obj_id,))
-                with lock_released:
-                    if not self.removal_threads:
-                        log.warn("remove(): no removal threads, removing synchronously")
-                        self._do_removal(obj_id)
-                    else:
+                if obj_id in self.in_transit:
+                    log.debug('remove(inode=%d, blockno=%d): deferring removal of '
+                              'object %d, still in transit', inode, blockno, obj_id)
+                    self.removed_in_transit.add(obj_id)
+                else:
+                    log.debug('remove(inode=%d, blockno=%d): deleting object %d',
+                              inode, blockno, obj_id)
+                    with lock_released:
                         self.to_remove.put(obj_id)
+                self.db.execute('DELETE FROM objects WHERE id=?', (obj_id,))
 
         log.debug('remove(inode=%d, start=%d, end=%s): end', inode, start_no, end_no)
 
@@ -741,12 +767,12 @@ class BlockCache(object):
         This method releases the global lock.
         """
 
-        for el in self.entries.itervalues():
+        for el in self.entries.values():
             if not (el.dirty and (el.inode, el.blockno) not in self.in_transit):
                 continue
 
             self.upload(el) # Releases global lock
-
+            
     def clear(self):
         """Clear cache
         
