@@ -6,19 +6,17 @@ Copyright (C) 2008-2009 Nikolaus Rath <Nikolaus@rath.org>
 This program can be distributed under the terms of the GNU GPLv3.
 '''
 
-from __future__ import division, print_function, absolute_import
-from .common import setup_logging, CTRL_NAME, QuietError
+from .logging import logging, setup_logging, QuietError
+from .common import assert_fs_owner, PICKLE_PROTOCOL
 from .parse_args import ArgumentParser
-import cPickle as pickle
-import errno
 import llfuse
-import logging
 import os
+import pickle
 import stat
 import sys
 import textwrap
 
-log = logging.getLogger("cp")
+log = logging.getLogger(__name__)
 
 def parse_args(args):
     '''Parse command line'''
@@ -38,6 +36,7 @@ def parse_args(args):
     parser.add_debug()
     parser.add_quiet()
     parser.add_version()
+    parser.add_fatal_warnings()
 
     parser.add_argument('source', help='source directory',
                         type=(lambda x: x.rstrip('/')))
@@ -78,24 +77,18 @@ def main(args=None):
     if fstat_p.st_dev != fstat_s.st_dev:
         raise QuietError('Source and target are not on the same file system.')
 
-    ctrlfile = os.path.join(parent, CTRL_NAME)
-    if not (CTRL_NAME not in llfuse.listdir(parent) and os.path.exists(ctrlfile)):
-        raise QuietError('Source and target are not on an S3QL file system')
-
-    if os.stat(ctrlfile).st_uid != os.geteuid() and os.geteuid() != 0:
-        raise QuietError('Only root and the mounting user may run s3qlcp.')
-
+    if os.path.ismount(options.source):
+        raise QuietError('%s is a mount point.' % options.source)
+        
+    ctrlfile = assert_fs_owner(options.source)
     try:
         os.mkdir(options.target)
-    except OSError as exc:
-        if exc.errno == errno.EACCES:
-            raise QuietError('No permission to create target directory')
-        else:
-            raise
+    except PermissionError:
+        raise QuietError('No permission to create target directory') from None
 
     fstat_t = os.stat(options.target)
     llfuse.setxattr(ctrlfile, 'copy', pickle.dumps((fstat_s.st_ino, fstat_t.st_ino),
-                                                    pickle.HIGHEST_PROTOCOL))
+                                                    PICKLE_PROTOCOL))
 
 
 if __name__ == '__main__':
