@@ -6,8 +6,6 @@ Copyright (C) 2008-2009 Nikolaus Rath <Nikolaus@rath.org>
 This program can be distributed under the terms of the GNU GPLv3.
 '''
 
-from __future__ import division, print_function, absolute_import 
-
 from llfuse import FUSEError
 from random import randint
 from s3ql import fs
@@ -20,14 +18,17 @@ from s3ql.metadata import create_tables
 from s3ql.database import Connection
 from s3ql.fsck import Fsck
 from s3ql.inode_cache import InodeCache
+from t2_block_cache import DummyQueue
+from common import catch_logmsg
 import errno
 import llfuse
 import os
 import shutil
 import stat
+import logging
 import tempfile
 import time
-import unittest2 as unittest
+import unittest
 
 # We need to access to protected members
 #pylint: disable=W0212
@@ -49,11 +50,12 @@ del stamp2
 class fs_api_tests(unittest.TestCase):
 
     def setUp(self):
-        self.backend_dir = tempfile.mkdtemp()
+        self.backend_dir = tempfile.mkdtemp(prefix='s3ql-backend-')
         plain_backend = local.Backend('local://' + self.backend_dir, None, None)
-        self.backend_pool = BackendPool(lambda: BetterBackend(b'schwubl', 'lzma', plain_backend))
+        self.backend_pool = BackendPool(lambda: BetterBackend(b'schwubl', ('zlib', 6),
+                                                              plain_backend))
         self.backend = self.backend_pool.pop_conn()
-        self.cachedir = tempfile.mkdtemp()
+        self.cachedir = tempfile.mkdtemp(prefix='s3ql-cache-')
         self.max_obj_size = 1024
 
         # Destructors are not guaranteed to run, and we can't unlink
@@ -69,23 +71,31 @@ class fs_api_tests(unittest.TestCase):
         # file system request handler
         llfuse.lock.acquire()
 
-        self.block_cache = BlockCache(self.backend_pool, self.db, self.cachedir + "/cache",
-                                      self.max_obj_size * 5)
-        self.server = fs.Operations(self.block_cache, self.db, self.max_obj_size,
+        cache = BlockCache(self.backend_pool, self.db, self.cachedir + "/cache",
+                           self.max_obj_size * 5)
+        self.block_cache = cache
+        self.server = fs.Operations(cache, self.db, self.max_obj_size,
                                     InodeCache(self.db, 0))
-
         self.server.init()
 
+        # Monkeypatch around the need for removal and upload threads
+        cache.to_remove = DummyQueue(cache)
+
+        class DummyDistributor:
+            def put(self, arg):
+                cache._do_upload(*arg)
+        cache.to_upload = DummyDistributor()
+        
         # Keep track of unused filenames
         self.name_cnt = 0
 
     def tearDown(self):
         self.server.inodes.flush()
+        llfuse.lock.release()
         self.block_cache.destroy()
         shutil.rmtree(self.cachedir)
         shutil.rmtree(self.backend_dir)
         os.unlink(self.dbfile.name)
-        llfuse.lock.release()
 
     @staticmethod
     def random_data(len_):
@@ -102,7 +112,7 @@ class fs_api_tests(unittest.TestCase):
 
     def newname(self):
         self.name_cnt += 1
-        return "s3ql_%d" % self.name_cnt
+        return ("s3ql_%d" % self.name_cnt).encode()
 
     def test_getattr_root(self):
         self.assertTrue(stat.S_ISDIR(self.server.getattr(ROOT_INODE).mode))
@@ -148,7 +158,7 @@ class fs_api_tests(unittest.TestCase):
 
         # Test with data in file
         fh = self.server.open(inode.id, os.O_RDWR)
-        self.server.write(fh, 0, 'foobar')
+        self.server.write(fh, 0, b'foobar')
         self.server.release(fh)
 
         self.server.extstat()
@@ -157,21 +167,21 @@ class fs_api_tests(unittest.TestCase):
 
     @staticmethod
     def dir_mode():
-        return (randint(0, 07777) & ~stat.S_IFDIR) | stat.S_IFDIR
+        return (randint(0, 0o7777) & ~stat.S_IFDIR) | stat.S_IFDIR
 
     @staticmethod
     def file_mode():
-        return (randint(0, 07777) & ~stat.S_IFREG) | stat.S_IFREG
+        return (randint(0, 0o7777) & ~stat.S_IFREG) | stat.S_IFREG
 
     def test_getxattr(self):
         (fh, inode) = self.server.create(ROOT_INODE, self.newname(),
                                          self.file_mode(), os.O_RDWR, Ctx())
         self.server.release(fh)
 
-        self.assertRaises(FUSEError, self.server.getxattr, inode.id, 'nonexistant-attr')
+        self.assertRaises(FUSEError, self.server.getxattr, inode.id, b'nonexistant-attr')
 
-        self.server.setxattr(inode.id, 'my-attr', 'strabumm!')
-        self.assertEqual(self.server.getxattr(inode.id, 'my-attr'), 'strabumm!')
+        self.server.setxattr(inode.id, b'my-attr', b'strabumm!')
+        self.assertEqual(self.server.getxattr(inode.id, b'my-attr'), b'strabumm!')
 
         self.server.forget([(inode.id, 1)])
         self.fsck()
@@ -212,11 +222,11 @@ class fs_api_tests(unittest.TestCase):
 
         self.assertListEqual([], self.server.listxattr(inode.id))
 
-        self.server.setxattr(inode.id, 'key1', 'blub')
-        self.assertListEqual(['key1'], self.server.listxattr(inode.id))
+        self.server.setxattr(inode.id, b'key1', b'blub')
+        self.assertListEqual([b'key1'], self.server.listxattr(inode.id))
 
-        self.server.setxattr(inode.id, 'key2', 'blub')
-        self.assertListEqual(sorted(['key1', 'key2']),
+        self.server.setxattr(inode.id, b'key2', b'blub')
+        self.assertListEqual(sorted([b'key1', b'key2']),
                              sorted(self.server.listxattr(inode.id)))
         self.server.forget([(inode.id, 1)])
         self.fsck()
@@ -244,7 +254,7 @@ class fs_api_tests(unittest.TestCase):
     def test_readdir(self):
 
         # Create a few entries
-        names = [ 'entry_%2d' % i for i in range(20) ]
+        names = [ ('entry_%2d' % i).encode() for i in range(20) ]
         for name in names:
             (fh, inode) = self.server.create(ROOT_INODE, name,
                                          self.file_mode(), os.O_RDWR, Ctx())
@@ -259,7 +269,7 @@ class fs_api_tests(unittest.TestCase):
 
         # Read all
         fh = self.server.opendir(ROOT_INODE)
-        self.assertListEqual(sorted(names + ['lost+found']) ,
+        self.assertListEqual(sorted(names + [b'lost+found']) ,
                              sorted(x[0] for x in self.server.readdir(fh, 0)))
         self.server.releasedir(fh)
 
@@ -277,7 +287,7 @@ class fs_api_tests(unittest.TestCase):
         except StopIteration:
             pass
 
-        self.assertListEqual(sorted(names + ['lost+found']) ,
+        self.assertListEqual(sorted(names + [b'lost+found']) ,
                              sorted(entries))
         self.server.releasedir(fh)
 
@@ -289,7 +299,7 @@ class fs_api_tests(unittest.TestCase):
         # Test that entries are deleted when they're no longer referenced
         (fh, inode) = self.server.create(ROOT_INODE, name,
                                          self.file_mode(), os.O_RDWR, Ctx())
-        self.server.write(fh, 0, 'foobar')
+        self.server.write(fh, 0, b'foobar')
         self.server.unlink(ROOT_INODE, name)
         self.assertFalse(self.db.has_val('SELECT 1 FROM contents JOIN names ON names.id = name_id '
                                          'WHERE name=? AND parent_inode = ?', (name, ROOT_INODE)))
@@ -306,9 +316,9 @@ class fs_api_tests(unittest.TestCase):
                                          self.file_mode(), os.O_RDWR, Ctx())
         self.server.release(fh)
 
-        self.assertRaises(FUSEError, self.server.removexattr, inode.id, 'some name')
-        self.server.setxattr(inode.id, 'key1', 'blub')
-        self.server.removexattr(inode.id, 'key1')
+        self.assertRaises(FUSEError, self.server.removexattr, inode.id, b'some name')
+        self.server.setxattr(inode.id, b'key1', b'blub')
+        self.server.removexattr(inode.id, b'key1')
         self.assertListEqual([], self.server.listxattr(inode.id))
         self.server.forget([(inode.id, 1)])
         self.fsck()
@@ -349,9 +359,9 @@ class fs_api_tests(unittest.TestCase):
 
         (fh, inode) = self.server.create(ROOT_INODE, oldname, self.file_mode(),
                                          os.O_RDWR, Ctx())
-        self.server.write(fh, 0, 'some data to deal with')
+        self.server.write(fh, 0, b'some data to deal with')
         self.server.release(fh)
-        self.server.setxattr(inode.id, 'test_xattr', '42*8')
+        self.server.setxattr(inode.id, b'test_xattr', b'42*8')
 
         inode_p_new = self.server.mkdir(ROOT_INODE, self.newname(), self.dir_mode(), Ctx())
         inode_p_new_before = self.server.getattr(inode_p_new.id).copy()
@@ -359,9 +369,9 @@ class fs_api_tests(unittest.TestCase):
 
         (fh, inode2) = self.server.create(inode_p_new.id, newname, self.file_mode(),
                                           os.O_RDWR, Ctx())
-        self.server.write(fh, 0, 'even more data to deal with')
+        self.server.write(fh, 0, b'even more data to deal with')
         self.server.release(fh)
-        self.server.setxattr(inode2.id, 'test_xattr', '42*8')
+        self.server.setxattr(inode2.id, b'test_xattr', b'42*8')
         self.server.forget([(inode2.id, 1)])
 
         time.sleep(CLOCK_GRANULARITY)
@@ -420,7 +430,7 @@ class fs_api_tests(unittest.TestCase):
         self.fsck()
 
     def test_setattr(self):
-        (fh, inode) = self.server.create(ROOT_INODE, self.newname(), 0641,
+        (fh, inode) = self.server.create(ROOT_INODE, self.newname(), 0o641,
                                          os.O_RDWR, Ctx())
         self.server.release(fh)
         inode_old = self.server.getattr(inode.id).copy()
@@ -439,7 +449,7 @@ class fs_api_tests(unittest.TestCase):
 
         for key in attr.__slots__:
             if getattr(attr, key) is not None:
-                self.assertEquals(getattr(attr, key),
+                self.assertEqual(getattr(attr, key),
                                   getattr(inode_new, key))
 
         self.server.forget([(inode.id, 1)])
@@ -489,8 +499,8 @@ class fs_api_tests(unittest.TestCase):
                                          self.file_mode(), os.O_RDWR, Ctx())
         self.server.release(fh)
 
-        self.server.setxattr(inode.id, 'my-attr', 'strabumm!')
-        self.assertEqual(self.server.getxattr(inode.id, 'my-attr'), 'strabumm!')
+        self.server.setxattr(inode.id, b'my-attr', b'strabumm!')
+        self.assertEqual(self.server.getxattr(inode.id, b'my-attr'), b'strabumm!')
         self.server.forget([(inode.id, 1)])
         self.fsck()
 
@@ -507,13 +517,13 @@ class fs_api_tests(unittest.TestCase):
                                          os.O_RDWR, Ctx())
         self.server.release(fh)
 
-        self.server.setxattr(inode.id, name1, 'strabumm!')
+        self.server.setxattr(inode.id, name1, b'strabumm!')
         self.fsck()
 
         self.server.removexattr(inode.id, name1)
         self.fsck()
 
-        self.server.setxattr(inode.id, name1, 'strabumm karacho!!')
+        self.server.setxattr(inode.id, name1, b'strabumm karacho!!')
         self.server.unlink(ROOT_INODE, name1)
         self.server.forget([(inode.id, 1)])
         self.fsck()
@@ -531,7 +541,7 @@ class fs_api_tests(unittest.TestCase):
 
         # Test with data in file
         fh = self.server.open(inode.id, os.O_RDWR)
-        self.server.write(fh, 0, 'foobar')
+        self.server.write(fh, 0, b'foobar')
         self.server.release(fh)
         self.server.forget([(inode.id, 1)])
         self.server.statfs()
@@ -561,11 +571,11 @@ class fs_api_tests(unittest.TestCase):
         name = self.newname()
 
         (fh, inode) = self.server.create(ROOT_INODE, name, self.file_mode(), os.O_RDWR, Ctx())
-        self.server.write(fh, 0, 'some data to deal with')
+        self.server.write(fh, 0, b'some data to deal with')
         self.server.release(fh)
 
         # Add extended attributes
-        self.server.setxattr(inode.id, 'test_xattr', '42*8')
+        self.server.setxattr(inode.id, b'test_xattr', b'42*8')
         self.server.forget([(inode.id, 1)])
 
         inode_p_before = self.server.getattr(ROOT_INODE).copy()
@@ -602,7 +612,7 @@ class fs_api_tests(unittest.TestCase):
     def test_relink(self):
         name = self.newname()
         name2 = self.newname()
-        data = 'some data to deal with'
+        data = b'some data to deal with'
 
         (fh, inode) = self.server.create(ROOT_INODE, name, self.file_mode(), os.O_RDWR, Ctx())
         self.server.write(fh, 0, data)
@@ -658,11 +668,13 @@ class fs_api_tests(unittest.TestCase):
         shutil.copy(datafile, datafile + '.bak')
         
         # Modify contents
-        with open(datafile, 'r+') as rfh:
+        with open(datafile, 'rb+') as rfh:
             rfh.seek(560)
             rfh.write(b'blrub!')
         with self.assertRaises(FUSEError) as cm:
-            self.server.read(fh, 0, len_)
+            with catch_logmsg('^Backend returned malformed data for',
+                              count=1, level=logging.ERROR):
+                self.server.read(fh, 0, len_)
         self.assertEqual(cm.exception.errno, errno.EIO)
         self.assertTrue(self.server.failsafe)
 
@@ -690,7 +702,9 @@ class fs_api_tests(unittest.TestCase):
         self.server.read(fh, 3, len_//2)
         self.server.cache.clear()
         with self.assertRaises(FUSEError) as cm:
-            self.server.read(fh, 5, len_//2)
+            with catch_logmsg('^Backend lost block',
+                              count=1, level=logging.ERROR):
+                self.server.read(fh, 5, len_//2)
         self.assertEqual(cm.exception.errno, errno.EIO)
         
         
@@ -732,56 +746,56 @@ class fs_api_tests(unittest.TestCase):
         self.fsck()
 
     def test_copy_tree(self):
-        ext_attr_name = 'system.foo.brazl'
-        ext_attr_val = 'schulla dku woumm bramp'
+        ext_attr_name = b'system.foo.brazl'
+        ext_attr_val = b'schulla dku woumm bramp'
 
-        src_inode = self.server.mkdir(ROOT_INODE, 'source', self.dir_mode(), Ctx())
-        dst_inode = self.server.mkdir(ROOT_INODE, 'dest', self.dir_mode(), Ctx())
+        src_inode = self.server.mkdir(ROOT_INODE, b'source', self.dir_mode(), Ctx())
+        dst_inode = self.server.mkdir(ROOT_INODE, b'dest', self.dir_mode(), Ctx())
 
         # Create file
-        (fh, f1_inode) = self.server.create(src_inode.id, 'file1',
+        (fh, f1_inode) = self.server.create(src_inode.id, b'file1',
                                             self.file_mode(), os.O_RDWR, Ctx())
-        self.server.write(fh, 0, 'file1 contents')
+        self.server.write(fh, 0, b'file1 contents')
         self.server.release(fh)
         self.server.setxattr(f1_inode.id, ext_attr_name, ext_attr_val)
 
         # Create hardlink
-        (fh, f2_inode) = self.server.create(src_inode.id, 'file2',
+        (fh, f2_inode) = self.server.create(src_inode.id, b'file2',
                                             self.file_mode(), os.O_RDWR, Ctx())
-        self.server.write(fh, 0, 'file2 contents')
+        self.server.write(fh, 0, b'file2 contents')
         self.server.release(fh)
-        f2_inode = self.server.link(f2_inode.id, src_inode.id, 'file2_hardlink')
+        f2_inode = self.server.link(f2_inode.id, src_inode.id, b'file2_hardlink')
 
         # Create subdirectory
-        d1_inode = self.server.mkdir(src_inode.id, 'dir1', self.dir_mode(), Ctx())
-        d2_inode = self.server.mkdir(d1_inode.id, 'dir2', self.dir_mode(), Ctx())
+        d1_inode = self.server.mkdir(src_inode.id, b'dir1', self.dir_mode(), Ctx())
+        d2_inode = self.server.mkdir(d1_inode.id, b'dir2', self.dir_mode(), Ctx())
 
         # ..with a 3rd hardlink
-        f2_inode = self.server.link(f2_inode.id, d1_inode.id, 'file2_hardlink')
+        f2_inode = self.server.link(f2_inode.id, d1_inode.id, b'file2_hardlink')
 
         # Replicate
         self.server.copy_tree(src_inode.id, dst_inode.id)
 
         # Change files
         fh = self.server.open(f1_inode.id, os.O_RDWR)
-        self.server.write(fh, 0, 'new file1 contents')
+        self.server.write(fh, 0, b'new file1 contents')
         self.server.release(fh)
 
         fh = self.server.open(f2_inode.id, os.O_RDWR)
-        self.server.write(fh, 0, 'new file2 contents')
+        self.server.write(fh, 0, b'new file2 contents')
         self.server.release(fh)
 
         # Get copy properties
-        f1_inode_c = self.server.lookup(dst_inode.id, 'file1')
-        f2_inode_c = self.server.lookup(dst_inode.id, 'file2')
-        f2h_inode_c = self.server.lookup(dst_inode.id, 'file2_hardlink')
-        d1_inode_c = self.server.lookup(dst_inode.id, 'dir1')
-        d2_inode_c = self.server.lookup(d1_inode_c.id, 'dir2')
-        f2_h_inode_c = self.server.lookup(d1_inode_c.id, 'file2_hardlink')
+        f1_inode_c = self.server.lookup(dst_inode.id, b'file1')
+        f2_inode_c = self.server.lookup(dst_inode.id, b'file2')
+        f2h_inode_c = self.server.lookup(dst_inode.id, b'file2_hardlink')
+        d1_inode_c = self.server.lookup(dst_inode.id, b'dir1')
+        d2_inode_c = self.server.lookup(d1_inode_c.id, b'dir2')
+        f2_h_inode_c = self.server.lookup(d1_inode_c.id, b'file2_hardlink')
 
         # Check file1
         fh = self.server.open(f1_inode_c.id, os.O_RDWR)
-        self.assertEqual(self.server.read(fh, 0, 42), 'file1 contents')
+        self.assertEqual(self.server.read(fh, 0, 42), b'file1 contents')
         self.server.release(fh)
         self.assertNotEqual(f1_inode.id, f1_inode_c.id)
         self.assertEqual(self.server.getxattr(f1_inode_c.id, ext_attr_name),
@@ -789,7 +803,7 @@ class fs_api_tests(unittest.TestCase):
 
         # Check file2
         fh = self.server.open(f2_inode_c.id, os.O_RDWR)
-        self.assertTrue(self.server.read(fh, 0, 42) == 'file2 contents')
+        self.assertTrue(self.server.read(fh, 0, 42) == b'file2 contents')
         self.server.release(fh)
         self.assertEqual(f2_inode_c.id, f2h_inode_c.id)
         self.assertEqual(f2_inode_c.refcount, 3)
@@ -799,18 +813,18 @@ class fs_api_tests(unittest.TestCase):
         # Check subdir1
         self.assertNotEqual(d1_inode.id, d1_inode_c.id)
         self.assertNotEqual(d2_inode.id, d2_inode_c.id)
-        self.server.forget(self.server.open_inodes.items())
+        self.server.forget(list(self.server.open_inodes.items()))
         self.fsck()
 
     def test_copy_tree_2(self):
-        src_inode = self.server.mkdir(ROOT_INODE, 'source', self.dir_mode(), Ctx())
-        dst_inode = self.server.mkdir(ROOT_INODE, 'dest', self.dir_mode(), Ctx())
+        src_inode = self.server.mkdir(ROOT_INODE, b'source', self.dir_mode(), Ctx())
+        dst_inode = self.server.mkdir(ROOT_INODE, b'dest', self.dir_mode(), Ctx())
 
         # Create file
-        (fh, inode) = self.server.create(src_inode.id, 'file1',
+        (fh, inode) = self.server.create(src_inode.id, b'file1',
                                      self.file_mode(), os.O_RDWR, Ctx())
-        self.server.write(fh, 0, 'block 1 contents')
-        self.server.write(fh, self.max_obj_size, 'block 1 contents')
+        self.server.write(fh, 0, b'block 1 contents')
+        self.server.write(fh, self.max_obj_size, b'block 1 contents')
         self.server.release(fh)
         self.server.forget([(inode.id, 1)])
 
@@ -821,23 +835,23 @@ class fs_api_tests(unittest.TestCase):
 
     def test_lock_tree(self):
 
-        inode1 = self.server.mkdir(ROOT_INODE, 'source', self.dir_mode(), Ctx())
+        inode1 = self.server.mkdir(ROOT_INODE, b'source', self.dir_mode(), Ctx())
 
         # Create file
-        (fh, inode1a) = self.server.create(inode1.id, 'file1',
+        (fh, inode1a) = self.server.create(inode1.id, b'file1',
                                             self.file_mode(), os.O_RDWR, Ctx())
-        self.server.write(fh, 0, 'file1 contents')
+        self.server.write(fh, 0, b'file1 contents')
         self.server.release(fh)
 
         # Create subdirectory
-        inode2 = self.server.mkdir(inode1.id, 'dir1', self.dir_mode(), Ctx())
-        (fh, inode2a) = self.server.create(inode2.id, 'file2',
+        inode2 = self.server.mkdir(inode1.id, b'dir1', self.dir_mode(), Ctx())
+        (fh, inode2a) = self.server.create(inode2.id, b'file2',
                                            self.file_mode(), os.O_RDWR, Ctx())
-        self.server.write(fh, 0, 'file2 contents')
+        self.server.write(fh, 0, b'file2 contents')
         self.server.release(fh)
 
         # Another file
-        (fh, inode3) = self.server.create(ROOT_INODE, 'file1',
+        (fh, inode3) = self.server.create(ROOT_INODE, b'file1',
                                           self.file_mode(), os.O_RDWR, Ctx())
         self.server.release(fh)
 
@@ -849,15 +863,15 @@ class fs_api_tests(unittest.TestCase):
 
         # Remove
         with self.assertRaises(FUSEError) as cm:
-            self.server._remove(inode1.id, 'file1', inode1a.id)
+            self.server._remove(inode1.id, b'file1', inode1a.id)
         self.assertEqual(cm.exception.errno, errno.EPERM)
 
         # Rename / Replace
         with self.assertRaises(FUSEError) as cm:
-            self.server.rename(ROOT_INODE, 'file1', inode1.id, 'file2')
+            self.server.rename(ROOT_INODE, b'file1', inode1.id, b'file2')
         self.assertEqual(cm.exception.errno, errno.EPERM)
         with self.assertRaises(FUSEError) as cm:
-            self.server.rename(inode1.id, 'file1', ROOT_INODE, 'file2')
+            self.server.rename(inode1.id, b'file1', ROOT_INODE, b'file2')
         self.assertEqual(cm.exception.errno, errno.EPERM)
 
         # Open
@@ -872,13 +886,13 @@ class fs_api_tests(unittest.TestCase):
         # Write
         fh = self.server.open(inode2a.id, os.O_RDONLY)
         with self.assertRaises(FUSEError) as cm:
-            self.server.write(fh, 0, 'foo')
+            self.server.write(fh, 0, b'foo')
         self.assertEqual(cm.exception.errno, errno.EPERM)
         self.server.release(fh)
 
         # Create
         with self.assertRaises(FUSEError) as cm:
-            self.server._create(inode2.id, 'dir1', self.dir_mode(), os.O_RDWR, Ctx())
+            self.server._create(inode2.id, b'dir1', self.dir_mode(), os.O_RDWR, Ctx())
         self.assertEqual(cm.exception.errno, errno.EPERM)
 
         # Setattr
@@ -888,39 +902,39 @@ class fs_api_tests(unittest.TestCase):
 
         # xattr
         with self.assertRaises(FUSEError) as cm:
-            self.server.setxattr(inode2.id, 'name', 'value')
+            self.server.setxattr(inode2.id, b'name', b'value')
         self.assertEqual(cm.exception.errno, errno.EPERM)
         with self.assertRaises(FUSEError) as cm:
-            self.server.removexattr(inode2.id, 'name')
+            self.server.removexattr(inode2.id, b'name')
         self.assertEqual(cm.exception.errno, errno.EPERM)
-        self.server.forget(self.server.open_inodes.items())
+        self.server.forget(list(self.server.open_inodes.items()))
         self.fsck()
 
     def test_remove_tree(self):
 
-        inode1 = self.server.mkdir(ROOT_INODE, 'source', self.dir_mode(), Ctx())
+        inode1 = self.server.mkdir(ROOT_INODE, b'source', self.dir_mode(), Ctx())
 
         # Create file
-        (fh, inode1a) = self.server.create(inode1.id, 'file1',
+        (fh, inode1a) = self.server.create(inode1.id, b'file1',
                                             self.file_mode(), os.O_RDWR, Ctx())
-        self.server.write(fh, 0, 'file1 contents')
+        self.server.write(fh, 0, b'file1 contents')
         self.server.release(fh)
 
         # Create subdirectory
-        inode2 = self.server.mkdir(inode1.id, 'dir1', self.dir_mode(), Ctx())
-        (fh, inode2a) = self.server.create(inode2.id, 'file2',
+        inode2 = self.server.mkdir(inode1.id, b'dir1', self.dir_mode(), Ctx())
+        (fh, inode2a) = self.server.create(inode2.id, b'file2',
                                            self.file_mode(), os.O_RDWR, Ctx())
-        self.server.write(fh, 0, 'file2 contents')
+        self.server.write(fh, 0, b'file2 contents')
         self.server.release(fh)
 
         # Remove
-        self.server.forget(self.server.open_inodes.items())
-        self.server.remove_tree(ROOT_INODE, 'source')
+        self.server.forget(list(self.server.open_inodes.items()))
+        self.server.remove_tree(ROOT_INODE, b'source')
 
-        for (id_p, name) in ((ROOT_INODE, 'source'),
-                             (inode1.id, 'file1'),
-                             (inode1.id, 'dir1'),
-                             (inode2.id, 'file2')):
+        for (id_p, name) in ((ROOT_INODE, b'source'),
+                             (inode1.id, b'file1'),
+                             (inode1.id, b'dir1'),
+                             (inode2.id, b'file2')):
             self.assertFalse(self.db.has_val('SELECT inode FROM contents JOIN names ON names.id = name_id '
                                              'WHERE name=? AND parent_inode = ?', (name, id_p)))
 
@@ -928,10 +942,3 @@ class fs_api_tests(unittest.TestCase):
             self.assertFalse(self.db.has_val('SELECT id FROM inodes WHERE id=?', (id_,)))
 
         self.fsck()
-
-
-def suite():
-    return unittest.makeSuite(fs_api_tests)
-
-if __name__ == "__main__":
-    unittest.main()
