@@ -6,8 +6,6 @@ Copyright (C) 2008-2009 Nikolaus Rath <Nikolaus@rath.org>
 This program can be distributed under the terms of the GNU GPLv3.
 '''
 
-from __future__ import division, print_function, absolute_import
-
 from s3ql.backends import local
 from s3ql.common import ROOT_INODE
 from s3ql.mkfs import init_tables
@@ -19,22 +17,18 @@ import shutil
 import stat
 import tempfile
 import time
-import unittest2 as unittest
+import unittest
 
 
 class fsck_tests(unittest.TestCase):
 
     def setUp(self):
-        self.backend_dir = tempfile.mkdtemp()
+        self.backend_dir = tempfile.mkdtemp(prefix='s3ql-backend-')
         self.backend = local.Backend('local://' + self.backend_dir, None, None)
-        self.cachedir = tempfile.mkdtemp()
+        self.cachedir = tempfile.mkdtemp(prefix='s3ql-cache-')
         self.max_obj_size = 1024
 
-        # Destructors are not guaranteed to run, and we can't unlink
-        # the file immediately because apsw refers to it by name. 
-        # Therefore, we unlink the file manually in tearDown() 
-        self.dbfile = tempfile.NamedTemporaryFile(delete=False)
-        
+        self.dbfile = tempfile.NamedTemporaryFile()
         self.db = Connection(self.dbfile.name)
         create_tables(self.db)
         init_tables(self.db)
@@ -46,7 +40,7 @@ class fsck_tests(unittest.TestCase):
     def tearDown(self):
         shutil.rmtree(self.cachedir)
         shutil.rmtree(self.backend_dir)
-        os.unlink(self.dbfile.name)
+        self.dbfile.close()
 
     def assert_fsck(self, fn):
         '''Check that fn detects and corrects an error'''
@@ -65,30 +59,30 @@ class fsck_tests(unittest.TestCase):
                                | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH,
                                os.getuid(), os.getgid(), time.time(), time.time(), time.time(),
                                1, 8))
-        self._link('test-entry', inode)
+        self._link(b'test-entry', inode)
 
         # Create new block
         fh = open(self.cachedir + '/%d-0' % inode, 'wb')
-        fh.write('somedata')
+        fh.write(b'somedata')
         fh.close()
         self.assert_fsck(self.fsck.check_cache)
-        self.assertEquals(self.backend['s3ql_data_1'], 'somedata')
+        self.assertEqual(self.backend['s3ql_data_1'], b'somedata')
 
         # Existing block
         self.db.execute('UPDATE inodes SET size=? WHERE id=?',
                         (self.max_obj_size + 8, inode))
         with open(self.cachedir + '/%d-1' % inode, 'wb') as fh:
-            fh.write('somedata')
+            fh.write(b'somedata')
         self.assert_fsck(self.fsck.check_cache)
 
         # Old block preserved
         with open(self.cachedir + '/%d-0' % inode, 'wb') as fh:
-            fh.write('somedat2')
+            fh.write(b'somedat2')
         self.assert_fsck(self.fsck.check_cache)
 
         # Old block removed
         with open(self.cachedir + '/%d-1' % inode, 'wb') as fh:
-            fh.write('somedat3')
+            fh.write(b'somedat3')
         self.assert_fsck(self.fsck.check_cache)
 
 
@@ -125,8 +119,8 @@ class fsck_tests(unittest.TestCase):
                               "VALUES (?,?,?,?,?,?,?,?)",
                               (stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR,
                                0, 0, time.time(), time.time(), time.time(), 1, 0))
-        self._link('name1', inode)
-        self._link('name2', inode)
+        self._link(b'name1', inode)
+        self._link(b'name2', inode)
         self.assert_fsck(self.fsck.check_inodes_refcount)
 
     def test_orphaned_inode(self):
@@ -143,22 +137,22 @@ class fsck_tests(unittest.TestCase):
                               "VALUES (?,?,?,?,?,?,?,?)",
                               (stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR,
                                0, 0, time.time(), time.time(), time.time(), 2, 0))
-        self._link('name1', inode)
-        self._link('name2', inode)
+        self._link(b'name1', inode)
+        self._link(b'name2', inode)
 
-        self.db.execute('UPDATE names SET refcount=refcount+1 WHERE name=?', ('name1',))
+        self.db.execute('UPDATE names SET refcount=refcount+1 WHERE name=?', (b'name1',))
 
         self.assert_fsck(self.fsck.check_names_refcount)
 
     def test_orphaned_name(self):
 
-        self._add_name('zupbrazl')
+        self._add_name(b'zupbrazl')
         self.assert_fsck(self.fsck.check_names_refcount)
 
     def test_contents_inode(self):
 
         self.db.execute('INSERT INTO contents (name_id, inode, parent_inode) VALUES(?,?,?)',
-                        (self._add_name('foobar'), 124, ROOT_INODE))
+                        (self._add_name(b'foobar'), 124, ROOT_INODE))
 
         self.assert_fsck(self.fsck.check_contents_inode)
 
@@ -169,7 +163,7 @@ class fsck_tests(unittest.TestCase):
                               (stat.S_IFDIR | stat.S_IRUSR | stat.S_IWUSR,
                                0, 0, time.time(), time.time(), time.time(), 1, 0))
         self.db.execute('INSERT INTO contents (name_id, inode, parent_inode) VALUES(?,?,?)',
-                        (self._add_name('foobar'), inode, 123))
+                        (self._add_name(b'foobar'), inode, 123))
 
         self.assert_fsck(self.fsck.check_contents_parent_inode)
 
@@ -211,12 +205,12 @@ class fsck_tests(unittest.TestCase):
                             "VALUES (?,?,?,?,?,?,?,?)",
                             (stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR,
                              0, 0, time.time(), time.time(), time.time(), 1, 128))
-        self._link('test-entry', id_)
+        self._link(b'test-entry', id_)
 
         obj_id = self.db.rowid('INSERT INTO objects (refcount,size) VALUES(?,?)', (1, 36))
         block_id = self.db.rowid('INSERT INTO blocks (refcount, obj_id, size) VALUES(?,?,?)',
                                  (1, obj_id, 512))
-        self.backend['s3ql_data_%d' % obj_id] = 'foo'
+        self.backend['s3ql_data_%d' % obj_id] = b'foo'
 
         # Case 1
         self.db.execute('UPDATE inodes SET size=? WHERE id=?', (self.max_obj_size + 120, id_))
@@ -243,7 +237,7 @@ class fsck_tests(unittest.TestCase):
 
     def test_objects_id(self):
         # Create an object that only exists in the backend
-        self.backend['s3ql_data_4364'] = 'Testdata'
+        self.backend['s3ql_data_4364'] = b'Testdata'
         self.assert_fsck(self.fsck.check_objects_id)
 
         # Create an object that does not exist in the backend
@@ -262,7 +256,7 @@ class fsck_tests(unittest.TestCase):
         self.db.execute('INSERT INTO inode_blocks (inode, blockno, block_id) VALUES(?,?,?)',
                         (id_, 0, block_id))
 
-        self._link('test-entry', id_)
+        self._link(b'test-entry', id_)
         self.assert_fsck(self.fsck.check_blocks_obj_id)
 
     def test_missing_obj(self):
@@ -278,14 +272,14 @@ class fsck_tests(unittest.TestCase):
         self.db.execute('INSERT INTO inode_blocks (inode, blockno, block_id) VALUES(?,?,?)',
                         (id_, 0, block_id))
 
-        self._link('test-entry', id_)
+        self._link(b'test-entry', id_)
         self.assert_fsck(self.fsck.check_objects_id)
 
 
     def test_inode_blocks_inode(self):
 
         obj_id = self.db.rowid('INSERT INTO objects (refcount, size) VALUES(1, 42)')
-        self.backend['s3ql_data_%d' % obj_id] = 'foo'
+        self.backend['s3ql_data_%d' % obj_id] = b'foo'
 
         block_id = self.db.rowid('INSERT INTO blocks (refcount, obj_id, size) VALUES(?,?,?)',
                                  (1, obj_id, 34))
@@ -304,7 +298,7 @@ class fsck_tests(unittest.TestCase):
         self.db.execute('INSERT INTO inode_blocks (inode, blockno, block_id) VALUES(?,?,?)',
                         (id_, 0, 35))
 
-        self._link('test-entry', id_)
+        self._link(b'test-entry', id_)
         self.assert_fsck(self.fsck.check_inode_blocks_block_id)
 
     def test_symlinks_inode(self):
@@ -317,7 +311,7 @@ class fsck_tests(unittest.TestCase):
     def test_ext_attrs_inode(self):
 
         self.db.execute('INSERT INTO ext_attributes (name_id, inode, value) VALUES(?,?,?)',
-                        (self._add_name('some name'), 34, b'some value'))
+                        (self._add_name(b'some name'), 34, b'some value'))
 
         self.assert_fsck(self.fsck.check_ext_attributes_inode)
 
@@ -327,7 +321,7 @@ class fsck_tests(unittest.TestCase):
                             "VALUES (?,?,?,?,?,?,?,?)",
                             (stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR,
                              0, 0, time.time(), time.time(), time.time(), 1, 128))
-        self._link('test-entry', id_)
+        self._link(b'test-entry', id_)
 
         self.db.execute('INSERT INTO ext_attributes (name_id, inode, value) VALUES(?,?,?)',
                         (34, id_, b'some value'))
@@ -352,7 +346,7 @@ class fsck_tests(unittest.TestCase):
         last = inodes[0]
         for inode in inodes[1:]:
             self.db.execute('INSERT INTO contents (name_id, inode, parent_inode) VALUES(?, ?, ?)',
-                            (self._add_name(bytes(inode)), inode, last))
+                            (self._add_name(str(inode).encode()), inode, last))
             last = inode
 
         self.assert_fsck(self.fsck.check_loops)
@@ -364,14 +358,14 @@ class fsck_tests(unittest.TestCase):
                                  (1, obj_id, 0))
         block_id_2 = self.db.rowid('INSERT INTO blocks (refcount, obj_id, size) VALUES(?,?,?)',
                                    (1, obj_id, 0))
-        self.backend['s3ql_data_%d' % obj_id] = 'foo'
+        self.backend['s3ql_data_%d' % obj_id] = b'foo'
 
         inode = self.db.rowid("INSERT INTO inodes (mode,uid,gid,mtime,atime,ctime,refcount,size) "
                               "VALUES (?,?,?,?,?,?,?,?)",
                               (stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR,
                                os.getuid(), os.getgid(), time.time(), time.time(), time.time(),
                                1, 2048))
-        self._link('test-entry', inode)
+        self._link(b'test-entry', inode)
         self.db.execute('INSERT INTO inode_blocks (inode, blockno, block_id) VALUES(?,?,?)',
                         (inode, 1, block_id_1))
         self.db.execute('INSERT INTO inode_blocks (inode, blockno, block_id) VALUES(?,?,?)',
@@ -387,7 +381,7 @@ class fsck_tests(unittest.TestCase):
     def test_wrong_block_refcount(self):
 
         obj_id = self.db.rowid('INSERT INTO objects (refcount, size) VALUES(1, 23)')
-        self.backend['s3ql_data_%d' % obj_id] = 'foo'
+        self.backend['s3ql_data_%d' % obj_id] = b'foo'
         block_id = self.db.rowid('INSERT INTO blocks (refcount, obj_id, size) VALUES(?,?,?)',
                                  (1, obj_id, 0))
 
@@ -395,7 +389,7 @@ class fsck_tests(unittest.TestCase):
                               "VALUES (?,?,?,?,?,?,?,?)",
                               (stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR, os.getuid(), os.getgid(),
                                time.time(), time.time(), time.time(), 1, self.max_obj_size))
-        self._link('test-entry', inode)
+        self._link(b'test-entry', inode)
 
         self.db.execute('INSERT INTO inode_blocks (inode, blockno, block_id) VALUES(?,?,?)',
                         (inode, 0, block_id))
@@ -407,7 +401,7 @@ class fsck_tests(unittest.TestCase):
     def test_orphaned_block(self):
 
         obj_id = self.db.rowid('INSERT INTO objects (refcount, size) VALUES(1, 24)')
-        self.backend['s3ql_data_%d' % obj_id] = 'foo'
+        self.backend['s3ql_data_%d' % obj_id] = b'foo'
         self.db.rowid('INSERT INTO blocks (refcount, obj_id, size) VALUES(?,?,?)',
                       (1, obj_id, 3))
         self.assert_fsck(self.fsck.check_blocks_refcount)
@@ -419,7 +413,7 @@ class fsck_tests(unittest.TestCase):
                         "VALUES (?,?,?,?,?,?,?,?,?)",
                         (inode, stat.S_IFIFO | stat.S_IRUSR | stat.S_IWUSR,
                          os.getuid(), os.getgid(), time.time(), time.time(), time.time(), 1, 0))
-        self._link('test-entry', inode)
+        self._link(b'test-entry', inode)
 
         self.fsck.found_errors = False
         self.fsck.check_unix()
@@ -433,14 +427,14 @@ class fsck_tests(unittest.TestCase):
     def test_unix_size_symlink(self):
 
         inode = 42
-        target = 'some funny random string'
+        target = b'some funny random string'
         self.db.execute("INSERT INTO inodes (id, mode,uid,gid,mtime,atime,ctime,refcount,size) "
                         "VALUES (?,?,?,?,?,?,?,?,?)",
                         (inode, stat.S_IFLNK | stat.S_IRUSR | stat.S_IWUSR,
                          os.getuid(), os.getgid(), time.time(), time.time(), time.time(), 1,
                          len(target)))
         self.db.execute('INSERT INTO symlink_targets (inode, target) VALUES(?,?)', (inode, target))
-        self._link('test-entry', inode)
+        self._link(b'test-entry', inode)
 
         self.fsck.found_errors = False
         self.fsck.check_unix()
@@ -457,7 +451,7 @@ class fsck_tests(unittest.TestCase):
                      "VALUES (?,?,?,?,?,?,?,?)",
                      (inode, stat.S_IFCHR | stat.S_IRUSR | stat.S_IWUSR,
                       os.getuid(), os.getgid(), time.time(), time.time(), time.time(), 1))
-        self._link('test-entry', inode)
+        self._link(b'test-entry', inode)
 
         self.fsck.found_errors = False
         self.fsck.check_unix()
@@ -474,7 +468,7 @@ class fsck_tests(unittest.TestCase):
         inode = self.db.rowid("INSERT INTO inodes (mode,uid,gid,mtime,atime,ctime,refcount) "
                               "VALUES (?,?,?,?,?,?,?)",
                               (perms, os.getuid(), os.getgid(), stamp, stamp, stamp, 1))
-        self._link('test-entry', inode)
+        self._link(b'test-entry', inode)
 
         self.assert_fsck(self.fsck.check_unix)
         
@@ -494,8 +488,8 @@ class fsck_tests(unittest.TestCase):
                               (perms | stat.S_IFREG, os.getuid(), os.getgid(), stamp, 
                                stamp, stamp, 1))
 
-        self._link('test-entry', inode)
-        self._link('subentry', inode2, inode)
+        self._link(b'test-entry', inode)
+        self._link(b'subentry', inode2, inode)
 
         self.assert_fsck(self.fsck.check_unix)
         
@@ -510,7 +504,7 @@ class fsck_tests(unittest.TestCase):
                               "VALUES (?,?,?,?,?,?,?)",
                               (stat.S_IFLNK | stat.S_IRUSR | stat.S_IWUSR,
                                os.getuid(), os.getgid(), time.time(), time.time(), time.time(), 1))
-        self._link('test-entry', inode)
+        self._link(b'test-entry', inode)
         self.fsck.check_unix()
         self.assertTrue(self.fsck.found_errors)
 
@@ -521,7 +515,7 @@ class fsck_tests(unittest.TestCase):
                      "VALUES (?,?,?,?,?,?,?,?)",
                      (inode, stat.S_IFIFO | stat.S_IRUSR | stat.S_IWUSR,
                       os.getuid(), os.getgid(), time.time(), time.time(), time.time(), 1))
-        self._link('test-entry', inode)
+        self._link(b'test-entry', inode)
 
         self.fsck.found_errors = False
         self.fsck.check_unix()
@@ -537,13 +531,13 @@ class fsck_tests(unittest.TestCase):
                    "VALUES (?,?,?,?,?,?,?)",
                    (stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR,
                     os.getuid(), os.getgid(), time.time(), time.time(), time.time(), 1))
-        self._link('test-entry', inode)
+        self._link(b'test-entry', inode)
 
         self.fsck.found_errors = False
         self.fsck.check_unix()
         self.assertFalse(self.fsck.found_errors)
         self.db.execute('INSERT INTO contents (name_id, inode, parent_inode) VALUES(?,?,?)',
-                        (self._add_name('foo'), ROOT_INODE, inode))
+                        (self._add_name(b'foo'), ROOT_INODE, inode))
         self.fsck.check_unix()
         self.assertTrue(self.fsck.found_errors)
 
@@ -553,7 +547,7 @@ class fsck_tests(unittest.TestCase):
                               "VALUES (?,?,?,?,?,?,?)",
                               (stat.S_IFSOCK | stat.S_IRUSR | stat.S_IWUSR,
                                os.getuid(), os.getgid(), time.time(), time.time(), time.time(), 1))
-        self._link('test-entry', inode)
+        self._link(b'test-entry', inode)
 
         self.fsck.found_errors = False
         self.fsck.check_unix()
@@ -568,12 +562,3 @@ class fsck_tests(unittest.TestCase):
 
         self.fsck.check_unix()
         self.assertTrue(self.fsck.found_errors)
-
-
-
-# Somehow important according to pyunit documentation
-def suite():
-    return unittest.makeSuite(fsck_tests)
-
-if __name__ == "__main__":
-    unittest.main()
